"use strict";
var check_1 = require("./check");
var containers_1 = require("./containers");
var keypath_1 = require("./keypath");
var string_1 = require("./string");
function deepSearch(object, keywords, selectors, stack) {
    containers_1.each(keypath_1.keyPaths(object), function (key) {
        var filtered = key;
        var unfiltered = key.split('.');
        var level = 0;
        for (var _i = 0, unfiltered_1 = unfiltered; _i < unfiltered_1.length; _i++) {
            var k = unfiltered_1[_i];
            if (select(keywords, k)) {
                var precedence = select(selectors, k);
                if (precedence > 0) {
                    filtered = filtered.replace(k + ".", '').replace("." + k, '');
                    level += precedence;
                }
                else {
                    // console.log('remove non match', filtered)
                    // unsetKeyPath(filtered, object);
                    return;
                }
            }
        }
        ;
        if (!stack[level]) {
            stack[level] = {};
        }
        stack[level][filtered] = keypath_1.valueForKeyPath(key, object);
    });
    var flat = {};
    containers_1.each(stack, function (level, height) {
        if (level) {
            containers_1.each(level, function (v, kp) {
                keypath_1.mergeValueAtKeypath(v, kp, flat);
            });
        }
    });
    return flat;
}
function shallowSearch(current, keywords, selectors, stack, height) {
    if (!check_1.check(stack[height], Object)) {
        stack[height] = {};
    }
    for (var _i = 0, _a = Object.keys(current); _i < _a.length; _i++) {
        var key = _a[_i];
        if (select(keywords, key)) {
            if (select(selectors, key)) {
                shallowSearch(current[key], keywords, selectors, stack, height + 1);
            }
        }
        else {
            if (check_1.check(current[key], Object)) {
                stack[height][key] =
                    flatten(shallowSearch(current[key], keywords, selectors, [], 0));
            }
            else {
                stack[height][key] = current[key];
            }
        }
    }
    return stack;
}
function flatten(stack) {
    var flat = {};
    for (var _i = 0, _a = Object.keys(stack); _i < _a.length; _i++) {
        var level = _a[_i];
        containers_1.extend(flat, stack[level]);
    }
    return flat;
}
function match(selectors, selectable) {
    if (string_1.startsWith(selectable, '!')) {
        return 1 * !containers_1.contains(selectors, selectable.slice(1));
    }
    return 1 * containers_1.contains(selectors, selectable);
}
function matchEvery(selectors, cssString) {
    if (cssString.indexOf(' ') !== -1) {
        var selectables = cssString.split(' ');
        return containers_1.sumIfEvery(selectables, function (selectable) {
            return match(selectors, selectable);
        });
    }
    return match(selectors, cssString);
}
function matchCssString(selectors, cssString) {
    var selectables = string_1.replaceAll(cssString, ', ', ',');
    if (selectables.indexOf(',') !== -1) {
        return containers_1.greatestResult(selectables.split(','), function (subCssString) {
            return matchEvery(selectors, subCssString);
        });
    }
    return matchEvery(selectors, selectables);
}
function select(input, cssString) {
    return matchCssString(input, cssString);
}
exports.select = select;
function search(tree, keywords, selectors, searchFn) {
    if (!tree) {
        throw new Error('searching undefined for selectors');
    }
    if (!keywords || !selectors) {
        console.warn('searching tree without keywords or selectors string');
    }
    return searchFn(tree, keywords, selectors, [], 0);
}
function cascadeShallow(tree, keywords, selectors) {
    return flatten(search(tree, keywords, selectors, shallowSearch));
}
exports.cascadeShallow = cascadeShallow;
function cascade(tree, keywords, selectors) {
    return search(tree, keywords, selectors, deepSearch);
}
exports.cascade = cascade;
