"use strict";
var check_1 = require("./check");
var decycle_1 = require("./decycle");
var pSlice = Array.prototype.slice;
function isArguments(object) {
    return Object.prototype.toString.call(object) == '[object Arguments]';
}
;
function each(iter, fn) {
    var broken = 0;
    var breakLoop = (function () { broken = 1; });
    if (check_1.check(iter, Array)) {
        var index = 0;
        for (var _i = 0, _a = iter; _i < _a.length; _i++) {
            var v = _a[_i];
            fn(v, index, breakLoop);
            if (broken) {
                return;
            }
            index++;
        }
    }
    if (check_1.check(iter, Object)) {
        for (var _b = 0, _c = Object.keys(iter); _b < _c.length; _b++) {
            var k = _c[_b];
            fn(iter[k], k, breakLoop);
            if (broken) {
                return;
            }
        }
    }
}
exports.each = each;
function extend(target) {
    var sources = [];
    for (var _i = 1; _i < arguments.length; _i++) {
        sources[_i - 1] = arguments[_i];
    }
    for (var _a = 0, sources_1 = sources; _a < sources_1.length; _a++) {
        var source = sources_1[_a];
        if (check_1.check(source, Object)) {
            for (var _b = 0, _c = Object.keys(source); _b < _c.length; _b++) {
                var key = _c[_b];
                if (check_1.check(source[key], Object) && check_1.check(target[key], Object)) {
                    extend(target[key], source[key]);
                }
                else {
                    target[key] = clone(source[key]);
                }
            }
        }
    }
    return target;
}
exports.extend = extend;
function combine() {
    var args = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        args[_i] = arguments[_i];
    }
    var result = {};
    for (var _a = 0, args_1 = args; _a < args_1.length; _a++) {
        var dict = args_1[_a];
        if (check_1.check(dict, Object)) {
            extend(result, dict);
        }
    }
    return result;
}
exports.combine = combine;
function any(iterable, fn) {
    for (var _i = 0, iterable_1 = iterable; _i < iterable_1.length; _i++) {
        var v = iterable_1[_i];
        if (fn(v) !== false) {
            return true;
        }
    }
    return false;
}
exports.any = any;
function every(iterable, fn) {
    for (var _i = 0, iterable_2 = iterable; _i < iterable_2.length; _i++) {
        var v = iterable_2[_i];
        if (fn(v) === false) {
            return false;
        }
    }
    return true;
}
exports.every = every;
function map(iter, fn) {
    var res = [];
    if (check_1.check(iter, Array)) {
        var i = 0;
        for (var _i = 0, _a = iter; _i < _a.length; _i++) {
            var v = _a[_i];
            res.push(fn(v, i));
            i++;
        }
    }
    if (check_1.check(iter, Object)) {
        for (var _b = 0, _c = Object.keys(iter); _b < _c.length; _b++) {
            var k = _c[_b];
            res.push(fn(iter[k], k));
        }
    }
    return res;
}
exports.map = map;
function reduce(input, fn, base) {
    var sum = base;
    each(input, function (value) {
        sum = fn(value, sum);
    });
    return sum;
}
exports.reduce = reduce;
function sum(input, fn) {
    var sum = 0;
    each(input, function (value) {
        sum = sum + fn(value);
    });
    return sum;
}
exports.sum = sum;
function greatestResult(input, fn) {
    var greatestResult = 0;
    each(input, function (value) {
        var res = fn(value);
        if (res > greatestResult)
            greatestResult = res;
    });
    return greatestResult;
}
exports.greatestResult = greatestResult;
function sumIfEvery(input, fn) {
    var sum = 0;
    each(input, function (value, index, breakLoop) {
        var res = fn(value);
        if (res > 0) {
            sum = sum + res;
        }
        else {
            sum = 0;
            breakLoop();
        }
    });
    return sum;
}
exports.sumIfEvery = sumIfEvery;
function geoSum(input, fn) {
    var sum = 1;
    each(input, function (value, key, breakLoop) {
        sum *= fn(value, sum);
    });
    return sum;
}
exports.geoSum = geoSum;
function union() {
    var args = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        args[_i] = arguments[_i];
    }
    var res = [];
    for (var _a = 0, args_2 = args; _a < args_2.length; _a++) {
        var arr = args_2[_a];
        for (var _b = 0, arr_1 = arr; _b < arr_1.length; _b++) {
            var v = arr_1[_b];
            if (!contains(res, v)) {
                res.push(v);
            }
        }
    }
    return res;
}
exports.union = union;
function intersect() {
    var args = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        args[_i] = arguments[_i];
    }
    var res = [];
    for (var _a = 0, args_3 = args; _a < args_3.length; _a++) {
        var a = args_3[_a];
        for (var _b = 0, a_1 = a; _b < a_1.length; _b++) {
            var v = a_1[_b];
            if (!contains(res, v)) {
                for (var _c = 0, args_4 = args; _c < args_4.length; _c++) {
                    var b = args_4[_c];
                    if (contains(b, v)) {
                        res.push(v);
                    }
                }
            }
        }
    }
    return res;
}
exports.intersect = intersect;
function difference(a, b) {
    var res = [];
    for (var _i = 0, a_2 = a; _i < a_2.length; _i++) {
        var v = a_2[_i];
        if (!contains(b, v)) {
            res.push(v);
        }
    }
    return res;
}
exports.difference = difference;
function contains(set, match) {
    if (check_1.check(match, Array)) {
        return containsAny(set, match);
    }
    for (var _i = 0, set_1 = set; _i < set_1.length; _i++) {
        var val = set_1[_i];
        if (isEqual(val, match)) {
            return true;
        }
    }
    return false;
}
exports.contains = contains;
function containsAny(set, match) {
    if (!check_1.check(match, Array)) {
        throw new Error('contains all takes a list to match');
    }
    for (var _i = 0, match_1 = match; _i < match_1.length; _i++) {
        var val = match_1[_i];
        if (contains(set, val)) {
            return true;
        }
    }
    return false;
}
exports.containsAny = containsAny;
function containsAll(set, match) {
    if (!check_1.check(match, Array)) {
        throw new Error('contains all takes a list to match');
    }
    for (var _i = 0, match_2 = match; _i < match_2.length; _i++) {
        var val = match_2[_i];
        if (!contains(set, val)) {
            return false;
        }
    }
    return true;
}
exports.containsAll = containsAll;
function isEqual(actual, expected, opts) {
    // http://wiki.commonjs.org/wiki/Unit_Testing/1.0
    if (!opts)
        opts = {};
    // 7.1. All identical values are equivalent, as determined by ===.
    if (actual === expected) {
        return true;
    }
    else if (actual instanceof Date && expected instanceof Date) {
        return actual.getTime() === expected.getTime();
    }
    else if (!actual || !expected || typeof actual != 'object' && typeof expected != 'object') {
        return opts.strict ? actual === expected : actual == expected;
    }
    else {
        return objEquiv(actual, expected, opts);
    }
}
exports.isEqual = isEqual;
function isUndefinedOrNull(value) {
    return value === null || value === undefined;
}
function isBuffer(x) {
    if (!x || typeof x !== 'object' || typeof x.length !== 'number')
        return false;
    if (typeof x.copy !== 'function' || typeof x.slice !== 'function') {
        return false;
    }
    if (x.length > 0 && typeof x[0] !== 'number')
        return false;
    return true;
}
function objEquiv(a, b, opts) {
    var i, key;
    if (isUndefinedOrNull(a) || isUndefinedOrNull(b))
        return false;
    // an identical 'prototype' property.
    if (a.prototype !== b.prototype)
        return false;
    //~~~I've managed to break Object.keys through screwy arguments passing.
    //   Converting to array solves the problem.
    if (isArguments(a)) {
        if (!isArguments(b)) {
            return false;
        }
        a = pSlice.call(a);
        b = pSlice.call(b);
        return isEqual(a, b, opts);
    }
    if (isBuffer(a)) {
        if (!isBuffer(b)) {
            return false;
        }
        if (a.length !== b.length)
            return false;
        for (i = 0; i < a.length; i++) {
            if (a[i] !== b[i])
                return false;
        }
        return true;
    }
    try {
        var ka = Object.keys(a), kb = Object.keys(b);
    }
    catch (e) {
        return false;
    }
    // having the same number of owned properties (keys incorporates
    // hasOwnProperty)
    if (ka.length != kb.length)
        return false;
    //the same set of keys (although not necessarily the same order),
    ka.sort();
    kb.sort();
    //~~~cheap key test
    for (i = ka.length - 1; i >= 0; i--) {
        if (ka[i] != kb[i])
            return false;
    }
    //equivalent values for every corresponding key, and
    //~~~possibly expensive deep test
    for (i = ka.length - 1; i >= 0; i--) {
        key = ka[i];
        if (!isEqual(a[key], b[key], opts))
            return false;
    }
    return typeof a === typeof b;
}
function _prune(input) {
    if (!check_1.check(input, Object)) {
        throw new Error('attempting to _prune undefined object');
    }
    var ref = input;
    var pruned = false;
    for (var _i = 0, _a = Object.keys(ref); _i < _a.length; _i++) {
        var k = _a[_i];
        var val = ref[k];
        if (check_1.check(val, Object)) {
            if (_prune(val)) {
                pruned = true;
            }
            if (isEmpty(val)) {
                delete ref[k];
                pruned = true;
            }
        }
    }
    return pruned;
}
function prune(obj) {
    _prune(obj);
    return obj;
}
exports.prune = prune;
function plain(obj) {
    return JSON.parse(JSON.stringify(obj));
}
exports.plain = plain;
function clone(input) {
    return decycle_1.retrocycle(decycle_1.decycle(input));
}
exports.clone = clone;
function arrayify(val) {
    if (check_1.check(val, Array)) {
        return val;
    }
    return [val];
}
exports.arrayify = arrayify;
function isEmpty(input) {
    var ref = input;
    if (!check_1.check(input, Object)) {
        throw new Error('checking non object for non-empty keys');
    }
    var containsValid = false;
    for (var _i = 0, _a = Object.keys(ref); _i < _a.length; _i++) {
        var k = _a[_i];
        if (check_1.check(ref[k], 'any')) {
            containsValid = true;
        }
    }
    return !containsValid;
}
function okmap(iterable, fn) {
    var sum = {};
    each(iterable, function (v, k) {
        var res = fn(v, k);
        var key = Object.keys(res)[0];
        sum[key] = res[key];
    });
    return sum;
}
exports.okmap = okmap;
function stringify(value, replacer, space) {
    return JSON.stringify(decycle_1.decycle(value), replacer, space || 2);
}
exports.stringify = stringify;
