"use strict";
var check_1 = require("./check");
var containers_1 = require("./containers");
var keypath_1 = require("./keypath");
function forwardDiffToModifier(prev, doc, fieldsToIgnore) {
    var filteredKeys = containers_1.union(containers_1.difference(keypath_1.keyPaths(prev), keypath_1.keyPaths(doc)), fieldsToIgnore || []);
    return diffToModifier(prev, doc, filteredKeys);
}
exports.forwardDiffToModifier = forwardDiffToModifier;
function shouldSet(val, prev) {
    if (Array.isArray(val)) {
        return !containers_1.isEqual(prev, val);
    }
    else if (val instanceof Date) {
        return !(prev instanceof Date) || (val.getTime() !== prev.getTime());
    }
    else if (check_1.check(val, Number)) {
        return (prev !== val) || !check_1.check(prev, Number);
    }
    else if (val !== null && typeof val === 'object') {
        return !containers_1.isEqual(prev, val);
    }
    else if (val) {
        return prev !== val;
    }
}
function shouldUnset(val, prev) {
    if ((prev || check_1.check(prev, Number)) && !(val || check_1.check(val, Number))) {
        return true;
    }
    if (val !== null && typeof prev === 'object') {
        if (!Object.keys(val).length) {
            return true;
        }
    }
    return false;
}
function diffToModifier(prev, doc, fieldsToIgnore, pruneEmptyObjects) {
    var delta = { $set: {}, $unset: {} };
    if (doc) {
        var forwardKeyPaths = keypath_1.filteredKeyPaths(keypath_1.keyPaths(doc), fieldsToIgnore || []);
        for (var _i = 0, forwardKeyPaths_1 = forwardKeyPaths; _i < forwardKeyPaths_1.length; _i++) {
            var keyPath = forwardKeyPaths_1[_i];
            var val = keypath_1.valueForKeyPath(keyPath, doc);
            if (shouldSet(val, keypath_1.valueForKeyPath(keyPath, prev))) {
                delta.$set[keyPath] = val;
            }
        }
    }
    if (prev) {
        var kps = keypath_1.keyPaths(prev, { allLevels: true });
        var existingKeyPaths = keypath_1.filteredKeyPaths(kps, fieldsToIgnore || []);
        for (var _a = 0, existingKeyPaths_1 = existingKeyPaths; _a < existingKeyPaths_1.length; _a++) {
            var keyPath = existingKeyPaths_1[_a];
            var curVal = keypath_1.valueForKeyPath(keyPath, doc);
            if (shouldUnset(curVal, keypath_1.valueForKeyPath(keyPath, prev))) {
                delta.$unset[keyPath] = true;
            }
        }
        var modifier = delta.$unset;
        var keys = Object.keys(modifier);
        for (var _b = 0, keys_1 = keys; _b < keys_1.length; _b++) {
            var pathA = keys_1[_b];
            for (var _c = 0, keys_2 = keys; _c < keys_2.length; _c++) {
                var pathB = keys_2[_c];
                if (keypath_1._keyPathContainsPath(pathA, pathB)) {
                    delete modifier[pathA];
                }
            }
        }
    }
    if (!Object.keys(delta.$set).length) {
        delete delta.$set;
    }
    if (!Object.keys(delta.$unset).length) {
        delete delta.$unset;
    }
    if (Object.keys(delta).length) {
        if (pruneEmptyObjects) {
            var newDelta = diffToModifier(prev, (containers_1.clone(prev), delta), fieldsToIgnore, false);
            return newDelta || delta;
        }
        return delta;
    }
}
exports.diffToModifier = diffToModifier;
function modifierToObj(modifier) {
    if (modifier) {
        var obj = {};
        for (var _i = 0, _a = Object.keys(modifier.$set || {}); _i < _a.length; _i++) {
            var keyPath = _a[_i];
            var val = modifier.$set[keyPath];
            keypath_1.setValueForKeyPath(val, keyPath, obj);
        }
        for (var _b = 0, _c = Object.keys(modifier.$unset || {}); _b < _c.length; _b++) {
            var keyPath = _c[_b];
            keypath_1.setValueForKeyPath(undefined, keyPath, obj);
        }
        return obj;
    }
}
exports.modifierToObj = modifierToObj;
function objToModifier(obj) {
    return diffToModifier(undefined, obj);
}
exports.objToModifier = objToModifier;
function apply(dest, source) {
    if (!source) {
        return dest;
    }
    if (source.$set || source.$unset) {
        $set(dest, source);
        $unset(dest, source);
    }
    else {
        var mod = objToModifier(source);
        $set(dest, mod);
        $unset(dest, mod);
    }
    return containers_1.prune(dest);
}
exports.apply = apply;
function $set(dest, source) {
    if (!source) {
        return;
    }
    if (source.$set || source.$unset) {
        $set(dest, source.$set);
    }
    return containers_1.each(source, function (val, keyPath) {
        if (check_1.check(val, Number) || val) {
            keypath_1.setValueForKeyPath(val, keyPath, dest);
        }
    });
}
exports.$set = $set;
function $addToSet(dest, src) {
    if (!Array.isArray(dest)) {
        throw new Error('$addToSet, 1st arg not array');
    }
    if (!containers_1.contains(dest, src)) {
        dest.push(src);
    }
    return dest;
}
exports.$addToSet = $addToSet;
function $unset(dest, source) {
    if (!source) {
        return;
    }
    if (source.$unset || source.$set) {
        $unset(dest, source.$unset);
    }
    containers_1.each(source, function (val, keyPath) { keypath_1.unsetKeyPath(keyPath, dest); });
}
exports.$unset = $unset;
function update(doc, options) {
    var model;
    if (check_1.check(options.get, Function)) {
        model = options.get();
    }
    else if (doc._id && options.collection) {
        model = options.collection.findOne({ _id: doc._id });
    }
    if (!model) {
        throw new Error('Diff: no doc to diff against');
    }
    var diff = diffToModifier(model, doc, options.ignore);
    if (diff) {
        if (!options.set && !options.collection) {
            throw new Error('Diff: no setter provided');
        }
        if (check_1.check(options.set, Function)) {
            var copy = containers_1.clone(model);
            apply(copy, diff);
            options.set(copy);
            if (!containers_1.isEqual(copy, model)) {
                throw new Error('Diff: not equal after update');
            }
        }
        else if (options.collection) {
            options.collection.update({ _id: model._id }, { $set: diff });
        }
    }
    return diff;
}
exports.update = update;
function mapModifierToKey(modifier, key) {
    if (!modifier) {
        throw new Error('called mapModifierToKey on undefined');
    }
    var valueModifier = {};
    for (var _i = 0, _a = Object.keys(modifier.$set || {}); _i < _a.length; _i++) {
        var keyPath = _a[_i];
        if (valueModifier.$set == null) {
            valueModifier.$set = {};
        }
        valueModifier.$set[key + "." + keyPath] = modifier.$set[keyPath];
    }
    for (var _b = 0, _c = Object.keys(modifier.$unset || {}); _b < _c.length; _b++) {
        var keyPath = _c[_b];
        if (valueModifier.$unset == null) {
            valueModifier.$unset = {};
        }
        valueModifier.$unset[key + "." + keyPath] = modifier.$set[keyPath];
    }
    return valueModifier;
}
exports.mapModifierToKey = mapModifierToKey;
