"use strict";
var check_1 = require("./check");
var containers_1 = require("./containers");
function setValueForKeyPath(value, keyPath, input) {
    var current = input;
    var keys = keyPath.split('.');
    for (var i = 0; i < keys.length - 1; i += 1) {
        var thisKey = keys[i];
        var nextKey = keys[i + 1];
        if (nextKey) {
            if (check_1.check(nextKey, Number)) {
                if (Array.isArray(current)) {
                    if (!Array.isArray(current[parseInt(thisKey, 10)])) {
                        current[parseInt(thisKey, 10)] = [];
                    }
                }
                else if (!Array.isArray(current[thisKey])) {
                    current[thisKey] = [];
                }
            }
            else if (Array.isArray(current)) {
                if (!(current[parseInt(thisKey, 10)] !== null &&
                    typeof current[parseInt(thisKey, 10)] === 'object')) {
                    current[parseInt(thisKey, 10)] = {};
                }
            }
            else if (!(current[thisKey] !== null &&
                typeof current[thisKey] === 'object')) {
                current[thisKey] = {};
            }
        }
        if (Array.isArray(current)) {
            current = current[parseInt(thisKey, 10)];
        }
        else {
            current = current[thisKey];
        }
    }
    var lastKey = keys[keys.length - 1];
    if (Array.isArray(current)) {
        current[parseInt(lastKey, 10)] = value;
    }
    else if (current !== null && typeof current === 'object') {
        current[lastKey] = value;
    }
}
exports.setValueForKeyPath = setValueForKeyPath;
function mergeValueAtKeypath(value, keyPath, obj) {
    // this function mutates obj
    var existing = valueForKeyPath(keyPath, obj);
    if (check_1.check(value, Object) && check_1.check(existing, Object)) {
        containers_1.extend(existing, value);
    }
    else if (check_1.check(value, Array) && check_1.check(existing, Array)) {
        setValueForKeyPath(containers_1.union(existing, value), keyPath, obj);
    }
    else {
        setValueForKeyPath(value, keyPath, obj);
    }
}
exports.mergeValueAtKeypath = mergeValueAtKeypath;
function valueForKeyPath(keyPath, input) {
    if (!input) {
        throw new Error('attempting to get valueForKeyPath on undefined object');
    }
    var current = input;
    var keys = keyPath.split('.');
    for (var i = 0; i < keys.length - 1; i += 1) {
        var key = keys[i];
        if (Array.isArray(current)) {
            if (!current[parseInt(key, 10)]) {
                return undefined;
            }
            current = current[parseInt(key, 10)];
        }
        else if (current !== null && typeof current === 'object') {
            if (!current[key]) {
                return undefined;
            }
            current = current[key];
        }
    }
    var lastKey = keys[keys.length - 1];
    if (Array.isArray(current)) {
        return current[parseInt(lastKey, 10)];
    }
    return current[lastKey];
}
exports.valueForKeyPath = valueForKeyPath;
function unsetKeyPath(keyPath, obj) {
    // this function mutates obj
    var keys = keyPath.split('.');
    var current = obj;
    for (var i = 0; i < keys.length - 1; i += 1) {
        var key = keys[i];
        if (Array.isArray(current)) {
            if (!current[parseInt(key, 10)]) {
                return 0;
            }
            current = current[parseInt(key, 10)];
        }
        else if (current !== null && typeof current === 'object') {
            if (!current[key]) {
                return 0;
            }
            current = current[key];
        }
    }
    var lastKey = keys[keys.length - 1];
    if (Array.isArray(current)) {
        var index = parseInt(lastKey, 10);
        if (current[index] !== undefined) {
            delete current[index];
            return 1;
        }
        return 0;
    }
    if (current[lastKey] !== undefined) {
        delete current[lastKey];
        return 1;
    }
    return 0;
}
exports.unsetKeyPath = unsetKeyPath;
function _keyPathContainsPath(keyPath, ignorePath) {
    var p = keyPath.split('.');
    var t = ignorePath.split('.');
    if (!(p.length > t.length)) {
        return false;
    }
    for (var i = 0; i < t.length; i += 1) {
        if (p[i] !== t[i]) {
            return false;
        }
    }
    return true;
}
exports._keyPathContainsPath = _keyPathContainsPath;
function keyPathContainsPath(keyPath, ignorePath) {
    if (keyPath === ignorePath) {
        return true;
    }
    return _keyPathContainsPath(keyPath, ignorePath);
}
exports.keyPathContainsPath = keyPathContainsPath;
function filteredKeyPaths(_keyPaths, ignore) {
    if (!ignore.length) {
        return _keyPaths;
    }
    var toFilter = [];
    for (var _i = 0, ignore_1 = ignore; _i < ignore_1.length; _i++) {
        var ignorePath = ignore_1[_i];
        for (var _a = 0, _keyPaths_1 = _keyPaths; _a < _keyPaths_1.length; _a++) {
            var keyPath = _keyPaths_1[_a];
            if (keyPathContainsPath(keyPath, ignorePath)) {
                toFilter.push(keyPath);
            }
        }
    }
    return containers_1.difference(_keyPaths, toFilter);
}
exports.filteredKeyPaths = filteredKeyPaths;
function keyPaths(obj, _options, _stack, parent) {
    var stack = _stack || [];
    var options = containers_1.clone(_options || {});
    var keys = Object.keys(obj);
    if (keys.length > 0) {
        for (var _i = 0, keys_1 = keys; _i < keys_1.length; _i++) {
            var el = keys_1[_i];
            var val = obj[el];
            if (Array.isArray(val)) {
                if (options.diffArrays) {
                    if (options.allLevels) {
                        stack.push(parent ? parent + "." + el : el);
                    }
                    for (var i = 0; i < val.length; i += 1) {
                        var p = parent ? parent + "." + el + "." + i : el + "." + i;
                        var s = val[i];
                        if (Array.isArray(s) || (s !== null && typeof s === 'object')) {
                            keyPaths(s, options, stack, p);
                        }
                        else {
                            stack.push(p);
                        }
                    }
                }
                else {
                    stack.push(parent ? parent + "." + el : el);
                }
            }
            else if (val instanceof Date) {
                var key = parent ? parent + "." + el : el;
                stack.push(key);
            }
            else if (val !== null && typeof val === 'object') {
                if (options.allLevels) {
                    stack.push(parent ? parent + "." + el : el);
                }
                keyPaths(val, options, stack, parent ? parent + "." + el : el);
            }
            else {
                stack.push(parent ? parent + "." + el : el);
            }
        }
    }
    else {
        stack.push(parent ? "" + parent : '');
    }
    return stack;
}
exports.keyPaths = keyPaths;
function allKeyPaths(obj) {
    return keyPaths(obj, { allLevels: true });
}
exports.allKeyPaths = allKeyPaths;
function flatObject(object, options) {
    var flat = {};
    var kpFunc = (options || {}).includeBranches ? allKeyPaths : keyPaths;
    for (var _i = 0, _a = kpFunc(object); _i < _a.length; _i++) {
        var keyPath = _a[_i];
        flat[keyPath] = valueForKeyPath(keyPath, object);
    }
    return flat;
}
exports.flatObject = flatObject;
