"use strict";
var check_1 = require("./check");
var containers_1 = require("./containers");
var graph_1 = require("./graph");
var OLHV = (function () {
    function OLHV() {
    }
    return OLHV;
}());
exports.OLHV = OLHV;
var OLHM = (function () {
    function OLHM() {
    }
    return OLHM;
}());
exports.OLHM = OLHM;
function parseOLHM(object) {
    var map = new OLHM();
    for (var _i = 0, _a = Object.keys(object); _i < _a.length; _i++) {
        var k = _a[_i];
        if (isOLHV(object(k))) {
            map[k] = object[k];
        }
        else {
            map[k] = {
                value: object[k]
            };
        }
    }
    return map;
}
exports.parseOLHM = parseOLHM;
function safeOLHM(olhm) {
    if (!olhm)
        return [];
    if (!check_1.check(olhm, Object)) {
        throw new Error('OLHM expects an object as input');
    }
    var keys = Object.keys(olhm);
    if (keys.length === 0) {
        return [];
    }
    // single key optimization
    if (keys.length === 1) {
        return [safeOLHV(olhm[keys[0]])];
    }
    // 2 or more keys, scan for dependencies
    var graph = new graph_1.Graph();
    for (var _i = 0, keys_1 = keys; _i < keys_1.length; _i++) {
        var k = keys_1[_i];
        graph.addNode(k);
    }
    for (var _a = 0, keys_2 = keys; _a < keys_2.length; _a++) {
        var k = keys_2[_a];
        if (isOLHV(olhm[k])) {
            graph.addDependency(k, olhm[k].require);
        }
    }
    // order based on dependencies
    var final = [];
    for (var _b = 0, _c = graph.overallOrder(); _b < _c.length; _b++) {
        var k = _c[_b];
        final.push(safeOLHV(olhm[k]));
    }
    return final;
}
exports.safeOLHM = safeOLHM;
function okmap(olhm, fn) {
    if (!check_1.check(olhm, Object)) {
        throw new Error('OLHM expects an object as input');
    }
    var keys = Object.keys(olhm);
    if (keys.length === 0) {
        return {};
    }
    var final = {};
    for (var _i = 0, keys_3 = keys; _i < keys_3.length; _i++) {
        var k = keys_3[_i];
        var olhv = olhm[k];
        if (isOLHV(olhv)) {
            var ret = fn(olhv.value, k);
            if (isOLHV(ret)) {
                final[k] = ret;
            }
            else {
                final[k] = {
                    value: ret,
                    require: olhv.require
                };
            }
        }
        else {
            final[k] = fn(olhv, k);
        }
    }
    return final;
}
exports.okmap = okmap;
function map(olhm, fn) {
    if (!olhm)
        return [];
    if (!check_1.check(olhm, Object)) {
        throw new Error('OLHM expects an object as input');
    }
    var keys = Object.keys(olhm);
    if (keys.length === 0) {
        return [];
    }
    // single key optimization
    if (keys.length === 1) {
        return [fn(safeOLHV(olhm[keys[0]]), keys[0])];
    }
    // 2 or more keys, scan for dependencies
    var graph = new graph_1.Graph();
    for (var _i = 0, keys_4 = keys; _i < keys_4.length; _i++) {
        var k = keys_4[_i];
        graph.addNode(k);
    }
    for (var _a = 0, keys_5 = keys; _a < keys_5.length; _a++) {
        var k = keys_5[_a];
        if (isOLHV(olhm[k])) {
            graph.addDependency(k, olhm[k].require);
        }
    }
    // order based on dependencies
    var final = [];
    for (var _b = 0, _c = graph.overallOrder(); _b < _c.length; _b++) {
        var k = _c[_b];
        var res = fn(safeOLHV(olhm[k]), k);
        final.push(res);
    }
    return final;
}
exports.map = map;
function reduce(olhm, fn, iv) {
    var iterable = safeOLHM(olhm);
    var i = 0;
    for (var _i = 0, iterable_1 = iterable; _i < iterable_1.length; _i++) {
        var v = iterable_1[_i];
        iv = fn(iv, v, i);
        i++;
    }
    return iv;
}
exports.reduce = reduce;
function isOLHV(obj) {
    if (check_1.check(obj, Object)) {
        var keys = Object.keys(obj);
        return keys.length === 2 && containers_1.containsAll(keys, ['require', 'value']);
    }
    return false;
}
exports.isOLHV = isOLHV;
function safeOLHV(objOrVal) {
    if (isOLHV(objOrVal)) {
        return objOrVal.value;
    }
    return objOrVal;
}
exports.safeOLHV = safeOLHV;
